(ns laser.rasterize)

(import '(java.io File))
(import '(org.apache.commons.io FileUtils))
(import '(javax.imageio ImageIO) )
(import '(javax.swing JFrame))
(import '(java.awt Color BorderLayout))
(import '(ij ImagePlus IJ))
(import '(java.lang Math))
(import '(java.awt Graphics2D Panel))
(import '(ij Macro))

(import '(java.io BufferedReader InputStreamReader))
(import '(java.awt.image BufferedImage))

(use 'clojure.contrib.str-utils)

(use 'clojure.contrib.combinatorics)


(use 'clojure.contrib.repl-utils)

(set! *print-length* 20)


(def img "/home/r/graster/test.png")


(def feed 120)
(def  dpi [500, 500])
(def on_range [0.0, 0.5])
(def overshoot 0.5)
(def offset [1.0, 1.0])
(def tiles [1, 1])
(def tile_size [false, false])
(def tile_spacing [0.125, 0.125])
(def feed 120)
(def cut_feed 20)
(def corner_radius 0)


(defn preserve-meta [f]
  (fn [& x] (with-meta
	      (apply f x)
	      (meta (last x)))))



(defn frame-hash
  "yields a convienent representation for the pixles in an image.
   Because of the size of the structvre generated, this must only be used
   in a transient way so that java can do it's garbage collection."
  [#^java.lang.String image-name]
  (let [image+ (ImagePlus. image-name)]
    (with-meta
      (let [buf (.. image+ getBufferedImage)
	    color (.getColorModel buf)]
	(apply hash-map 
	       (interleave 
		(doall (for [x (range (.getWidth image+)) y (range (.getHeight image+))]
			 (vector x y)))
		(doall (for [x (range (.getWidth image+)) y (range (.getHeight image+))]
			 (let [data (.getRGB buf x y)] 
			   (hash-map :r (bit-shift-right (bit-and 0xff0000 data) 16)
				     :g (bit-shift-right (bit-and 0x00ff00 data) 8)
				     :b (bit-and 0x0000ff data))))))))
      {:width  (.getWidth image+) :height (.getHeight image+)})))







(def white {:r 255, :g 255, :b 255})
(def black {:r 0,   :g 0,   :b 0})

(def expt #(Math/pow %1 %2))

(defn rgb-euclidian
  [{r1 :r g1 :g b1 :b} {r2 :r g2 :g b2 :b} ]
  (expt (+ (expt (- r1 r2) 2)
	  (expt (- g1 g2) 2)
	  (expt (- b1 b2) 2)) 0.5))

(defn b&w
  "turn everything strictly black or white"
  [window]
  (with-meta
  (zipmap
   (keys window)
   (map (fn  [rgb] 
	  (if (> (rgb-euclidian rgb white) (rgb-euclidian rgb black))
	    black white))
	(vals window))) (meta window)))



(defn raster-preamble []
  (str-join \newline
	    ["M63 P0\nG61"
	    (str \F feed)
	    "M101"
	    "M3 S1"]))

(defn raster-epilogue []
  (str-join \newline
	    ["M63 PO"
	     "M5"
	     "M2"]))


(defn raster-comment [string]
  (str "(" (re-gsub #"[()]" "" string) ")"))

(defn filter-keys [fun m]
  (select-keys m (filter fun (keys m))))
(def filter-keys (preserve-meta filter-keys))

(defn filter-vals [fun m]
    (into {} (filter (comp fun val) m)))
(def filter-vals (preserve-meta filter-vals))

(defn frame-hash->bufferedImage
  [frame-hash]
    (let [data  (meta frame-hash)
	image (BufferedImage. (:width data) (:height data) BufferedImage/TYPE_INT_BGR)]
	
	(doall (for [element frame-hash] 
	  (let [coord  (key element) 
		rgb   (val element)
		packed-RGB 
		(+ (bit-shift-left (:r rgb) 16)
		   (bit-shift-left (:g rgb) 8)
		                   (:b rgb))]
	    (.setRGB image (first coord) (last coord) packed-RGB))))
	image))
  
(defmulti  display "Creates a JFrame and displays a buffered image"  class)

(defn- makePanel [image] (proxy [Panel] [] (paint [g]  (.drawImage g image 0 0 nil))))


(defmethod display 
  BufferedImage  [image] 
  (let [panel (makePanel image)
	frame (JFrame. "Oh Yeah!")]
    (.add frame panel) 
    (.pack frame) 
    (.setVisible frame true ) 
    (.setSize frame(.getWidth image) (.getHeight image))))
 
(defmethod display
  ImagePlus [image]
  (display (.getBufferedImage image)))

(defmethod display
  clojure.lang.PersistentHashMap [frame-hash]
  (display (frame-hash->bufferedImage frame-hash)))

(defmethod display
    clojure.lang.PersistentArrayMap [frame-hash]
    (display (frame-hash->bufferedImage frame-hash)))





